unit MainForm;

interface

uses NFormW, NCanvasW, NConfW, NCommand,
  NkVImage, ImgFilter, SpiFilter, InfoDialog, ShellUtils, FileList;

var
  Config: TConf;
  ProfDir: WideString;

type
  TNkVMainForm = class(TForm)
    ArFileList: TFileList;
    Archiv: TArchive;
    BgColor: Longint;
    ContextMenu: TContextMenu;
    Command: TCommand;
    DirMode: Boolean;
    DispWidth, DispHeight: Integer;
    FileList: TFileList;
    FileListUpdate: Boolean;
    FitImage: Boolean;
    FitLargeImage: Boolean;
    GridSize: Integer;
    Image: TNkVImage;
    InfoDialog: TInfoDialog;
    OldPt: TPoint;
    OrgLeft, OrgTop: Integer;
    Plugins: TPlugins;
    PluginsDir: WideString;
    ScaleFilter: Boolean;
    ShowColorValue: Boolean;
    ShowGrid: Boolean;
    UsePluginListFile: Boolean;
    ViewLeft, ViewTop: Integer;
    ViewWidth, ViewHeight: Integer;
    ZoomValue: Double;
    constructor Create; override;
    destructor Destroy; override;
    procedure Closing; override;
    procedure DefaultHandler(var Message); override;
    procedure DispatchCommand(Id: Longint); override;
    procedure DropFiles(const ListDropFiles: TListDropFiles); override;
    procedure MouseDoubleClick(const Button: TMouseButton; CX, CY: Longint);
        override;
    procedure MouseDown(const Button: TMouseButton; CX, CY: Longint); override;
    procedure MouseMove(CX, CY: Longint); override;
    procedure MouseUp(const Button: TMouseButton; CX, CY: Longint); override;
    procedure MouseWheel(Delta: Longint); override;
    procedure Paint(const Canvas: TCanvas; Left, Top, Width, Height: Longint);
        override;
    procedure RecvFromRemote(const S: WideString); override;
    procedure Resize; override;
    procedure ShowContextMenu(SX, SY: Longint); override;

    procedure Zoom(Value: Double);
    procedure Scroll(DX, DY: Integer);
    procedure LoadPluginList;
    procedure LoadFirstFile(const List: TFileList);
    procedure LoadFile(const Path: WideString; Last: Boolean);
    procedure LoadFirstFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadLastFileFromArchive(const Ar: TArchive;
      const List: TFileList);
    procedure LoadFileFromArchive(const Ar: TArchive; const FileName: string);
    procedure NextImage(const List: TFileList);
    procedure PrevImage(const List: TFileList; Last: Boolean);
    procedure GetImageCoord(Cx, Cy: Integer; var X, Y: Integer);
    procedure GetColorValue(Ix, Iy: Integer; var R, G, B: Integer);
    procedure UpdateDispSize;
    procedure UpdateCaption;
    procedure UpdateCursor;
    procedure UpdateInfo;
    procedure UpdateWinSize;
    procedure FillFileList;
    procedure Error(const Path: WideString; const Msg: WideString);
    procedure SetWallpaper(Tile: Boolean);
    procedure ShowPluginConfig(const PluginFileName: string);
  published
    procedure cmAbout;
    procedure cmAboutPlugins;
    procedure cmCopyColorValue;
    procedure cmCopyFileName;
    procedure cmExit;
    procedure cmFirstFile;
    procedure cmFitImage;
    procedure cmFitLargeImage;
    procedure cmFitWindow;
    procedure cmFullScreen;
    procedure cmHideBorder;
    procedure cmInfo;
    procedure cmLastFile;
    procedure cmNextFile;
    procedure cmNextImage;
    procedure cmNextWindow;
    procedure cmOpenDir;
    procedure cmPrevFile;
    procedure cmPrevImage;
    procedure cmPrevWindow;
    procedure cmRefreshFileList;
    procedure cmReloadImage;
    procedure cmSaveAs;
    procedure cmScaleFilter;
    procedure cmScrollBottom;
    procedure cmScrollDown;
    procedure cmScrollLeft;
    procedure cmScrollRight;
    procedure cmScrollTop;
    procedure cmScrollTopBottom;
    procedure cmScrollUp;
    procedure cmSetImageSize;
    procedure cmSetWallpaper;
    procedure cmSetWallpaperTile;
    procedure cmShowColorValue;
    procedure cmShowGrid;
    procedure cmStayOnTop;
    procedure cmZoom100;
    procedure cmZoom150;
    procedure cmZoom200;
    procedure cmZoom300;
    procedure cmZoom400;
    procedure cmZoom50;
    procedure cmZoom800;
    procedure cmZoomIn;
    procedure cmZoomOut;
    procedure cmPluginMenu;
  end;

implementation

uses NTypes, NLibW, NSysW, NStrW, NClipboard,
    Menu, ifdib, ifpnm, ifgdiplus, AboutDialog, ImageSizeDialog, PluginsDialog;

const
  ZoomTable: array [0..18] of Double = (
    0.0833, 0.125, 0.167, 0.25, 0.333, 0.5, 0.667,
    1.0, 1.25, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 12.0, 16.0
  );
  KeyScroll = 16;
  IDM_PLUGINSETTING_FIRST = 4000;
  IDM_PLUGINSETTING_LAST = IDM_PLUGINSETTING_FIRST + 999;

constructor TNkVMainForm.Create;
begin
  inherited;

  Command := TCommand.Create(Self, [
    cmAbout,
    cmAboutPlugins,
    cmCopyColorValue,
    cmCopyFileName,
    cmExit,
    cmFirstFile,
    cmFitImage,
    cmFitLargeImage,
    cmFitWindow,
    cmFullScreen,
    cmHideBorder,
    cmInfo,
    cmLastFile,
    cmNextFile,
    cmNextImage,
    cmNextWindow,
    cmOpenDir,
    cmPrevFile,
    cmPrevImage,
    cmPrevWindow,
    cmRefreshFileList,
    cmReloadImage,
    cmSaveAs,
    cmScaleFilter,
    cmScrollBottom,
    cmScrollDown,
    cmScrollLeft,
    cmScrollRight,
    cmScrollTop,
    cmScrollTopBottom,
    cmScrollUp,
    cmSetImageSize,
    cmSetWallpaper,
    cmSetWallpaperTile,
    cmShowColorValue,
    cmShowGrid,
    cmStayOnTop,
    cmZoom100,
    cmZoom150,
    cmZoom200,
    cmZoom300,
    cmZoom400,
    cmZoom50,
    cmZoom800,
    cmZoomIn,
    cmZoomOut,
    cmPluginMenu
  ]);

  SetText(App.Name);
  SetBackground(-1);
  EnableDoubleClick(True);
  EnableDropFiles(True);

  ZoomValue := 1.0;

  Image := TNkVImage.Create;
  InfoDialog := TInfoDialog.Create;
  FileList := TFileList.Create;
  ArFileList := TFileList.Create;
 
  // set border before loading position to correct size
  if Config.HasKey('hideborder') then
    SetBorder(not Config.GetBool('hideborder'));
  if App.AlreadyExists then // window position is default
  begin
   Config.Del('x');
   Config.Del('y');
  end;
  LoadFormPos(Self, Config);
  FitImage := Config.GetBool('fitimage');
  FitLargeImage := Config.GetBool('fitlargeimage');
  ScaleFilter := Config.GetBool('scalefilter');
  if Config.HasKey('background') then
    BgColor := Config.GetInt('background', 0, 0) shl 16 or
      Config.GetInt('background', 1, 0) shl 8 or
      Config.GetInt('background', 2, 0)
  else
    BgColor := SysColors.Window;
  PluginsDir := Config.GetStrW('pluginsdir', 0, DirName(ProgPath));
  PluginsDir := AbsPath(JoinPath([ProfDir, PluginsDir]));
  GridSize := Config.GetInt('gridsize', 0);

  LoadPluginList;
end;

destructor TNkVMainForm.Destroy;
begin
  FileList.Free;
  ArFileList.Free;
  InfoDialog.Free;
  Command.Free;
end;

procedure TNkVMainForm.DefaultHandler(var Message);
begin
  if Assigned(ContextMenu) then
    ContextMenu.HandleMenuMsg(Message);
  inherited;
end;

procedure TNkVMainForm.Closing;
var
  position: TConf;
begin
  position := TConf.Create;
  try
    SaveFormPos(Self, position);
    try
      position.SaveToFile(JoinPath([ProfDir, 'position']));
    except
      on E: EIOError do App.Error(E.Path+EOL+EOL+E.Msg);
    end;
  finally
    position.Free;
  end;
end;

procedure TNkVMainForm.LoadPluginList;
  function InUseExtInfo(const Path: string): Boolean;
  var
    i: Integer;
    args: TStringArray;
    name: string;
  begin
    args := Config['usepluginextinfo'];
    if args = nil then
    begin
      Result := False;
      Exit;
    end;

    name := StrUpper(BaseName(Path));
    for i := 0 to High(args) do
      if StrUpper(args[i]) = name then
      begin
        Result := True;
        Exit;
      end;
    Result := False;
  end;
var
  i: Integer;
  line, tail: string;
  dir: TDir;
  f: TFile;
begin
  SetLength(Plugins, 32);
  i := 0;
  try
    f := TFile.Open(JoinPath([ProfDir, 'pluginlist']));
    try
      while TConf.ReadLine(f, line) do
      begin
        if i = Length(Plugins) then
          SetLength(Plugins, i + i div 4);
        Plugins[i].Path := line;
        Plugins[i].UseExtInfo := InUseExtInfo(Plugins[i].Path);
        Inc(i);
      end;
    finally
      f.Free;
    end;
    UsePluginListFile := True;
  except
    on E: EIOError do
      try
        with TListDir.Create(PluginsDir) do
          try
            while Each(dir) do
            begin
              if dmDirectory in dir.Mode then
                Continue;
              tail := Copy(dir.Name, Length(dir.Name)-3, 4);
              if (tail = '.spi') or (tail = '.SPI') then
              begin
                if i = Length(Plugins) then
                  SetLength(Plugins, i + i div 4);
                Plugins[i].Path := JoinPath([PluginsDir, dir.Name]);
                Plugins[i].UseExtInfo := InUseExtInfo(Plugins[i].Path);
                Inc(i);
              end;
            end;
          finally
            Free;
          end;
      except
        on E: EIOError do App.Error('plugins directory: '+E.Path+EOL+EOL+E.Msg);
    end;
  end;
  SetLength(Plugins, i);
end;

procedure TNkVMainForm.FillFileList;
var
  s, path: WideString;
  i: Integer;
begin
  if not DirMode then
    Exit;

  if not FileListUpdate then
    Exit;

  FileListUpdate := False;
  s := FileList.Items[FileList.Cur];
  path := DirName(s);
  FileList.Clear;
  try
    FileList.AddDir(path);
  except
    on E: EIOError do Error(E.Path, E.Msg);
  end;
  FileList.Sort;

  for i := 0 to FileList.Count-1 do 
    if FileList.Items[i] = s then
    begin
      FileList.Cur := i;
      Exit;
    end;
end;

procedure TNkVMainForm.LoadFirstFile(const List: TFileList);
var
  i: Integer;
  multi, stdin: Boolean;
begin
  multi := List.Count > 1;
  stdin := (List.Count > 0) and (List.Items[0] = '');
  while List.Count > 0 do
  begin
    try
      LoadFile(List.Items[0], False)
    except
      on E: EIOError do
      begin
        if not stdin then
          Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        if not multi then
          Error(List.Items[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;

  if List.Count > 0 then // loding succeeded
  begin
    FileList.Clear;
    for i := 0 to List.Count - 1 do
      FileList.Add(List.Items[i]);
    FileList.Cur := 0;
    ViewLeft := 0;
    ViewTop := 0;
    DispWidth := RoundOff(Image.Width * ZoomValue);
    DispHeight := RoundOff(Image.Height * ZoomValue);
    UpdateDispSize;
    UpdateWinSize;
    UpdateInfo;
    UpdateCursor;
  end
  else if not stdin and multi then
    Error('', 'Loadable file not found.');
end;

procedure TNkVMainForm.UpdateCursor;
var
  cursorimage: TCursorImage;
begin
  if ShowColorValue then
    cursorimage := Cursors.Cross
  else
    cursorimage := Cursors.Arrow;
  SetCursor(cursorimage);
  Cursor.SetImage(cursorimage);
end;

procedure TNkVMainForm.LoadFileFromArchive(const Ar: TArchive;
  const FileName: string);
var
  buf: PChar;
  fsize: Cardinal;
  filter: TObject;
  oldcur: TCursorImage;
begin
  oldcur := Cursor.SetImage(Cursors.Wait);
  fsize := Ar[FileName].FileSize;
  GetMem(buf, fsize);
  try
    Ar.Extract(FileName, buf, fsize);
    filter := LookupFilter(FileName, buf, fsize, Plugins);
    try
      if not Assigned(filter) or (filter is TArchive) then
        raise EImgFilterError.Create('Unknown format.');
      Image.ReadMem(buf, fsize, filter as TImgFilter);
    finally
      filter.Free;
    end;
  finally
    FreeMem(buf);
    Cursor.SetImage(oldcur);
  end;
end;

procedure TNkVMainForm.UpdateWinSize;
begin
  if not FitImage and not FitLargeImage and not FullScreen and
      not Maximized then
    SetClientSize(DispWidth, DispHeight);
  Update;
end;

procedure TNkVMainForm.LoadFirstFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List.ByteItems[0]);
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(0);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[0], E.Msg);
        List.Delete(0);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := 0;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadLastFileFromArchive(const Ar: TArchive;
  const List: TFileList);
begin
  while List.Count > 0 do
  begin
    try
      LoadFileFromArchive(Ar, List.ByteItems[List.Count-1])
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
      on E: EImgFilterError do
      begin
        //Error(List[List.Count-1], E.Msg);
        List.Delete(List.Count-1);
        Continue;
      end;
    end;
    Break;
  end;
  if List.Count > 0 then // loding succeeded
  begin
    Archiv.Free;
    Archiv := Ar;
    ArFileList.Free;
    ArFileList := List;
    ArFileList.Cur := List.Count-1;
  end else
    raise EImgFilterError.Create('Loadable file not found in archive.');
end;

procedure TNkVMainForm.LoadFile(const Path: WideString; Last: Boolean);
const
  TestBufSize = 2048;
  StdinBufSize = 32768;
var
  len, fsize: Cardinal;
  buf: PChar;
  filter: TObject;
  f: TFile;
  isstdin: Boolean;
  oldcur: TCursorImage;
  ar: TArchive;
  arlist: TFileList;
  fname: string;
begin
  isstdin := Path = '';
  if isstdin then
    f := TFile.CreateStdIn
  else
    f := TFile.Open(Path);

  oldcur := Cursor.SetImage(Cursors.Wait);
  buf := nil;
  try
    GetMem(buf, TestBufSize);
    FillChar(buf^, TestBufSize, 0);
    len := f.Read(buf^, TestBufSize);
    if isstdin and (len = 0) then
      raise EIOError.Create('<STDIN>');
    filter := LookupFilter(Path, buf, len, Plugins);
    try
      if not Assigned(filter) then
        raise EImgFilterError.Create('Unknown format.');

      if filter is TArchive then
      begin
        ar := filter as TArchive;
        try
          ar.Open(Path);
        except
          on E: EImgFilterError do
          begin
            filter.Free;
            filter := nil;
            raise;
          end;
        end;

        arlist := TFileList.Create;
        while ar.EachName(fname) do
          if ar.Info[fname].FileSize > 0 then
            arlist.Add(fname);
        arlist.Sort;
        if Last then
          LoadLastFileFromArchive(ar, arlist)
        else
          LoadFirstFileFromArchive(ar, arlist);
      end
      else begin
        if isstdin then
        begin
          // stdin
          fsize := len;
          while True do
          begin
            ReallocMem(buf, fsize + StdinBufSize);
            try
              len := f.Read(buf[fsize], StdinBufSize);
            except
              on E: EIOError do Break;
            end;
            if len = 0 then
              Break;
            Inc(fsize, len);
          end;
        end else
        begin
          // normal file
          fsize := f.Size;
          ReallocMem(buf, fsize);
          f.Read(buf[len], fsize - len);
        end; 
        Image.ReadMem(buf, fsize, filter as TImgFilter);
        ArFileList.Clear;
      end;
    finally
      if not (filter is TArchive) then
        filter.Free;
    end;
  finally
    FreeMem(buf);
    f.Free;
    Cursor.SetImage(oldcur);
  end;
end;

procedure TNkVMainForm.Zoom(Value: Double);
begin
  if Image.Loaded then
  begin
    if ZoomValue > 0 then
    begin
      ViewLeft := RoundOff((ViewLeft+ViewWidth/2)/ZoomValue*Value-ViewWidth/2);
      ViewTop := RoundOff((ViewTop+ViewHeight/2)/ZoomValue*Value-ViewHeight/2);
      if ViewLeft < 0 then ViewLeft := 0;
      if ViewTop < 0 then ViewTop := 0;
    end;
    FitImage := False;
    FitLargeImage := False;
    DispWidth := RoundOff(Image.Width * Value);
    DispHeight := RoundOff(Image.Height * Value);
    UpdateDispSize;
  end;
end;

procedure TNkVMainForm.Scroll(DX, DY: Longint);
var
  cw, ch, vw, vh: Integer;
begin
  if not Image.Loaded then
    Exit;

  if FitImage or FitLargeImage then
    Exit;

  GetClientSize(cw, ch);
  if DispWidth < cw then
    vw := DispWidth
  else
    vw := cw;
  if DispHeight < ch then
    vh := DispHeight
  else
    vh := ch;

  Inc(ViewLeft, DX);
  if ViewLeft < 0 then
  begin
    Dec(DX, ViewLeft);
    ViewLeft := 0;
  end
  else if ViewLeft > (DispWidth-vw) then
  begin
    Inc(DX, DispWidth-vw-ViewLeft);
    ViewLeft := DispWidth-vw;
  end;

  Inc(ViewTop, DY);
  if ViewTop < 0 then
  begin
    Dec(DY, ViewTop);
    ViewTop := 0;
  end
  else if ViewTop > (DispHeight-vh) then
  begin
    Inc(DY, DispHeight-vh-ViewTop);
    ViewTop := DispHeight-vh;
  end;

  inherited Scroll(DX, DY);
  Update;
end;

procedure TNkVMainForm.NextImage(const List: TFileList);
var
  i: Integer;
begin
  i := list.Cur+1;
  while i < list.Count do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list.ByteItems[i])
      else
        LoadFile(list.Items[i], False)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        list.Delete(i);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Continue;
      end;
    end;
    Break;
  end;
  if i < list.Count then
    list.Cur := i;
  ViewLeft := 0;
  ViewTop := 0;
  DispWidth := RoundOff(Image.Width * ZoomValue);
  DispHeight := RoundOff(Image.Height * ZoomValue);
  UpdateDispSize;
  UpdateWinSize;
  UpdateInfo;
end;

procedure TNkVMainForm.PrevImage(const List: TFileList; Last: Boolean);
var
  i: Integer;
begin
  i := list.Cur-1;
  while i >= 0 do
  begin
    try
      if list = ArFileList then
        LoadFileFromArchive(Archiv, list.ByteItems[i])
      else
        LoadFile(list.Items[i], Last)
    except
      on E: EIOError do
      begin
        Error(E.Path, E.Msg);
        FileList.Delete(i);
        Dec(i);
        Dec(FileList.Cur);
        Continue;
      end;
      on E: EImgFilterError do // XXX
      begin
        list.Delete(i);
        Dec(i);
        Dec(list.Cur);
        Continue;
      end;
    end;
    Break;
  end;
  if i >= 0 then
    list.Cur := i;
  ViewLeft := 0;
  ViewTop := 0;
  DispWidth := RoundOff(Image.Width * ZoomValue);
  DispHeight := RoundOff(Image.Height * ZoomValue);
  UpdateDispSize;
  UpdateWinSize;
  UpdateInfo;
end;

procedure TNkVMainForm.UpdateInfo;
var
  dir: TDir;
begin
  InfoDialog.Path := AbsPath(FileList.Items[FileList.Cur]);
  if ArFileList.Count = 0 then
  begin
    InfoDialog.PathInArchive := '';
    InfoDialog.ArchivePluginFileName := '';
    InfoDialog.ArchivePluginName := '';
    if InfoDialog.Path <> '' then
    begin
      try
        dir := DirStat(InfoDialog.Path);
        InfoDialog.MTime := dir.MTime;
        InfoDialog.ATime := dir.ATime;
        InfoDialog.DirStatError := False;
      except
        on E: EIOError do
        begin
          Error(E.Path, E.Msg);
          InfoDialog.DirStatError := True;
        end;
      end;
    end
  end
  else begin
    InfoDialog.PathInArchive := ArFileList.Items[ArFileList.Cur];
    InfoDialog.MTime := Archiv[InfoDialog.PathInArchive].DateTime;
    InfoDialog.ArchivePluginFileName := Archiv.PluginFileName;
    InfoDialog.ArchivePluginName := Archiv.PluginName;
    InfoDialog.DirStatError := False;
  end;
  InfoDialog.Image := Image;

  if InfoDialog.Visible then
    InfoDialog.UpdateInfo;
end;

procedure TNkVMainForm.GetImageCoord(Cx, Cy: Integer; var X, Y: Integer);
begin
  X := Trunc((Cx + ViewLeft-OrgLeft) / ZoomValue);
  Y := Trunc((Cy + ViewTop-OrgTop) / ZoomValue);
end;

procedure TNkVMainForm.GetColorValue(Ix, Iy: Integer; var R, G, B: Integer);
begin
  if (Ix<Image.Width) and (Iy<Image.Height) and (Ix>=0) and (Iy>=0) then
    Image.GetPixel(Ix, Iy, R, G, B)
  else begin
    R := -1;
    G := -1;
    B := -1;
  end;
end;

procedure TNkVMainForm.UpdateCaption;
var
  fname: WideString;
  info, mode: string;
  pt: TPoint;
  r, g, b, x, y, ix, iy: Integer;
begin
  if not Image.Loaded then
    Exit;

  if FileList.Items[FileList.Cur] = '' then
    fname := '<STDIN>'
  else
    fname := BaseName(FileList.Items[FileList.Cur]);

  if ArFileList.Count > 0 then
    fname := fname + '/' + ArFileList.Items[ArFileList.Cur];

  if ShowColorValue then
  begin
    pt := Cursor.GetPosition;
    ScreenToClient(pt.x, pt.y, x, y);
    GetImageCoord(x, y, ix, iy);
    GetColorValue(ix, iy, r, g, b);
    if r > -1 then
      info := Format(
        '[R{0:03} G{1:03} B{2:03} #{0:02x}{1:02x}{2:02x}] (X{3} Y{4})',
        [r, g, b, ix, iy])
    else
      info := '[R--- G--- B---] (X--- Y---)';
  end
  else begin
    if Image.Bpp > 0 then
      info := Format('({0}x{1}x{2})', [Image.Width, Image.Height, Image.Bpp])
    else
      info := Format('({0}x{1}x?)', [Image.Width, Image.Height]);
  end;

  if FitImage then
    mode := ' (Fit Image Mode)'
  else if FitLargeImage then
    mode := ' (Fit Large Image Mode)'
  else
    mode := '';

  SetText(fname + Format('  {0} {1}%{2} - {3}', [info,
    IStr(RoundOff(ZoomValue*100)), mode, App.Name]));
end;

procedure TNkVMainForm.Error(const Path: WideString; const Msg: WideString);
begin
  if Path = '' then
    inherited Error(Msg)
  else
    inherited Error(Path + EOL + EOL + Msg);
end;

procedure TNkVMainForm.SetWallpaper(Tile: Boolean);
var
  fname: string;
begin
  if Image.Loaded then
  begin
    fname := JoinPath([ProfDir, 'wallpaper.bmp']);
    try
      Image.Write(fname, Image.Width, Image.Height, False);
      ShellUtils.SetWallpaper(fname, Tile);
    except
      on E: EIOError do Error(E.Path, E.Msg);
    end;
  end;
end;

procedure TNkVMainForm.UpdateDispSize;
var
  cw, ch: Integer;
begin
  if not Image.Loaded then
    Exit;
  if Minimized then
    Exit;

  GetClientSize(cw, ch);
  if FitImage or
     (FitLargeImage and ((Image.Width > cw) or (Image.Height > ch))) then
  begin
    ViewLeft := 0;
    ViewTop := 0;
    if Image.Width * ch > Image.Height * cw then
    begin
      DispWidth := cw;
      DispHeight := RoundOff(Image.Height * DispWidth / Image.Width);
    end else
    begin
      DispHeight := ch;
      DispWidth := RoundOff(Image.Width * DispHeight / Image.Height);
    end;
  end
  else if FitLargeImage then
  begin
    ViewLeft := 0;
    ViewTop := 0;
    DispWidth := Image.Width;
    DispHeight := Image.Height;
  end;

  if DispWidth < cw then
    ViewWidth := DispWidth 
  else
    ViewWidth := cw;
  if DispHeight < ch then
    ViewHeight := DispHeight 
  else
    ViewHeight := ch;

  if ViewLeft > (DispWidth-ViewWidth) then
    ViewLeft := DispWidth-ViewWidth;
  if ViewTop > (DispHeight-ViewHeight) then
    ViewTop := DispHeight-ViewHeight;

  if ViewWidth < cw then
    OrgLeft := (cw - ViewWidth) div 2
  else
    OrgLeft := 0;
  if ViewHeight < ch then
    OrgTop := (ch - ViewHeight) div 2
  else
    OrgTop := 0;

  ZoomValue := DispWidth / Image.Width;

  UpdateCaption;
  Invalidate;
end;

{ Command Methods }

procedure TNkVMainForm.cmStayOnTop;
begin
  SetStayOnTop(not StayOnTop)
end;

procedure TNkVMainForm.cmHideBorder;
begin
  if not FullScreen then 
    SetBorder(not Border);
  UpdateCursor;
end;
  
procedure TNkVMainForm.cmFitImage;
begin
  FitImage := not FitImage;
  if FitImage and FitLargeImage then
    FitLargeImage := False;
  UpdateDispSize;
end;

procedure TNkVMainForm.cmFitLargeImage;
begin
  FitLargeImage := not FitLargeImage;
  if FitImage and FitLargeImage then
    FitImage := False;
  UpdateDispSize;
end;

procedure TNkVMainForm.cmScaleFilter;
begin
  ScaleFilter := not ScaleFilter;
  Invalidate;
end;

procedure TNkVMainForm.cmShowGrid;
begin
  ShowGrid := not ShowGrid;
  Invalidate;
end;

procedure TNkVMainForm.cmNextWindow;
begin
  App.ActivateNext(False);
end;

procedure TNkVMainForm.cmPrevWindow;
begin
  App.ActivateNext(True);
end;

procedure TNkVMainForm.cmAboutPlugins;
var
  i: Integer;
  s: WideString;
begin
  if Length(Plugins) > 0 then
  begin
    s := '';
    for i := 0 to High(Plugins) do
      with TSpiFilter.Create(nil) do
        try
          LoadPlugin(Plugins[i].Path, Plugins[i].UseExtInfo);
          if Handle = 0 then
            s := s + '(not found)'#13#10 + Plugins[i].Path
          else begin
            s := s + Name + #13#10 + FileName;
            // -1: trim last ';'
            if Plugins[i].UseExtInfo then
              s := s + #13#10 + Copy(ExtInfo, 1, Length(ExtInfo)-1);
          end;
          s := s + #13#10#13#10;
        finally
          Free;
        end;
  end else
    s := 'No plugins.';
  with TPluginsDialog.Create do
    try
      Text := s;
      ShowModal(Self);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.cmAbout;
begin
  with TAboutDialog.Create do
    try
      ProfDir := MainForm.ProfDir;
      if UsePluginListFile then
        PluginsDir := '(use pluginlist file)'
      else
        PluginsDir := Self.PluginsDir;
      ShowModal(Self);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.cmExit;
begin
  Close;
end;

procedure TNkVMainForm.cmFitWindow;
begin
  if Image.Loaded and not FullScreen then
  begin
    if Maximized then
      Restore;
    SetClientSize(DispWidth, DispHeight);
  end;
end;

procedure TNkVMainForm.cmFullScreen;
begin
  if Image.Loaded then
    SetFullScreen(not FullScreen, False);
end;

procedure TNkVMainForm.cmZoomIn;
var
  i: Integer;
begin
  if Image.Loaded then
    for i := 0 to High(ZoomTable) do
      if DispWidth < RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
end;

procedure TNkVMainForm.cmZoomOut;
var
  i: Integer;
begin
  if Image.Loaded then
    for i := High(ZoomTable) downto 0 do
      if DispWidth > RoundOff(Image.Width * ZoomTable[i]) then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
end;

procedure TNkVMainForm.cmZoom50; begin Zoom(0.5); end;
procedure TNkVMainForm.cmZoom100; begin Zoom(1.0); end;
procedure TNkVMainForm.cmZoom150; begin Zoom(1.5); end;
procedure TNkVMainForm.cmZoom200; begin Zoom(2.0); end;
procedure TNkVMainForm.cmZoom300; begin Zoom(3.0); end;
procedure TNkVMainForm.cmZoom400; begin Zoom(4.0); end;
procedure TNkVMainForm.cmZoom800; begin Zoom(8.0); end;
  
procedure TNkVMainForm.cmSetWallpaper;
begin
  SetWallpaper(False);
end;

procedure TNkVMainForm.cmSetWallpaperTile;
begin
  SetWallpaper(True);
end;

procedure TNkVMainForm.cmScrollLeft;   begin Scroll(-KeyScroll, 0) end;
procedure TNkVMainForm.cmScrollRight;  begin Scroll(KeyScroll, 0) end;
procedure TNkVMainForm.cmScrollUp;     begin Scroll(0, -KeyScroll) end;
procedure TNkVMainForm.cmScrollDown;   begin Scroll(0, KeyScroll) end;
procedure TNkVMainForm.cmScrollTop;    begin Scroll(0, -DispHeight) end;
procedure TNkVMainForm.cmScrollBottom; begin Scroll(0, DispHeight) end;

procedure TNkVMainForm.cmScrollTopBottom;
var
  w, h: Integer;
begin
  GetClientSize(w, h);
  if Viewtop = DispHeight - h then
    Scroll(0, -DispHeight)
  else
    Scroll(0, DispHeight);
end;

procedure TNkVMainForm.cmShowColorValue;
begin
  if Image.Loaded then
  begin
    ShowColorValue := not ShowColorValue;
    UpdateCaption;
    UpdateCursor;
  end;
end;

procedure TNkVMainForm.cmFirstFile;
begin
  if Image.Loaded then
  begin
    if FileList.Items[FileList.Cur] = '' then  // stdin
      Exit;
    FillFileList;
    FileList.Cur := -1;
    NextImage(FileList);
  end;
end;

procedure TNkVMainForm.cmLastFile;
begin
  if Image.Loaded then
  begin
    if FileList.Items[FileList.Cur] = '' then  // stdin
      Exit;
    FillFileList;
    FileList.Cur := FileList.Count;
    PrevImage(FileList, False);
  end;
end;

procedure TNkVMainForm.cmNextFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    NextImage(FileList);
  end;
end;

procedure TNkVMainForm.cmOpenDir;
var
  path: WideString;
begin
  if Image.Loaded then
    try
      path := AbsPath(FileList.Items[FileList.Cur]);
      OpenDir(path);
    except
      on E: EOSError do
        Error(path, E.Msg);
    end;
end;

procedure TNkVMainForm.cmPrevFile;
begin
  if Image.Loaded then
  begin
    FillFileList;
    PrevImage(FileList, False);
  end;
end;

procedure TNkVMainForm.cmNextImage;
begin
  if Image.Loaded then
    if (ArFileList.Count = 0) or (ArFileList.Cur = ArFilelist.Count-1) then
    begin
      FillFileList;
      NextImage(FileList)
    end else
      NextImage(ArFileList);
end;

procedure TNkVMainForm.cmPrevImage;
begin
  if Image.Loaded then
    if (ArFileList.Count = 0) or (ArFileList.Cur = 0) then
    begin
      FillFileList;
      PrevImage(FileList, True)
    end else
      PrevImage(ArFileList, True);
end;

procedure TNkVMainForm.cmCopyFileName;
var
  s: string;
begin
  if Image.Loaded then
    with TClipboard.Create do
      try
        s := FileList.Items[FileList.Cur];
        if ArFileList.Count > 0 then
          s := s + '/' + ArFileList.Items[ArFileList.Cur];
        Write(s[1], Length(s));
      finally
        Free;
      end;
end;

procedure TNkVMainForm.cmCopyColorValue;
var
  r, g, b, x, y, ix, iy: Integer;
  pt: TPoint;
  s: string;
begin
  if Image.Loaded then
  begin
    pt := Cursor.GetPosition;
    ScreenToClient(pt.x, pt.y, x, y);
    GetImageCoord(x, y, ix, iy);
    GetColorValue(ix, iy, r, g, b);
    if r > -1 then
      with TClipboard.Create do
        try
          s := Format('#{0:02x}{1:02x}{2:02x}', [r, g, b]);
          Write(s[1], Length(s));
        finally
          Free;
        end;
  end;
end;

procedure TNkVMainForm.cmSaveAs;
var
  fname: WideString;
begin
  if not Image.Loaded then
    Exit;

  if FileList.Items[FileList.Cur] = '' then
    fname := 'stdin'
  else if ArFileList.Count > 0 then
    fname := BaseName(ArFileList.Items[ArFileList.Cur])
  else
    fname := BaseName(FileList.Items[FileList.Cur]);
  fname := fname + '.bmp';

  if ShowSaveDlg(Handle, ['All Files', '*.*'], fname) then
    try
      Image.Write(fname, DispWidth, DispHeight, ScaleFilter);
    except
      on E: EIOError do Error(E.Path, E.Msg);
    end;
end;

procedure TNkVMainForm.cmInfo;
begin
  if Image.Loaded and not InfoDialog.Visible then
  begin
    InfoDialog.ShowModeless(Self);
    UpdateInfo;
  end;
end;

procedure TNkVMainForm.cmReloadImage;
begin
  if Image.Loaded then
  begin
    LoadFile(FileList.Items[FileList.Cur], False);
    UpdateDispSize;
    UpdateWinSize;
    UpdateInfo;
  end;
end;

procedure TNkVMainForm.cmRefreshFileList;
begin
  if Image.Loaded and DirMode then
    FileListUpdate := True; 
end;

procedure TNkVMainForm.cmPluginMenu; begin end; // dummy

procedure TNkVMainForm.cmSetImageSize;
begin
  if Image.Loaded then
    with TImageSizeDlg.Create do
      try
        Image := Self.Image;
        Width := DispWidth;
        Height := DispHeight;
        ShowModal(Self);
        if Update then
        begin
          FitImage := False;
          FitLargeImage := False;
          DispWidth := Width;
          DispHeight := Height;
          UpdateDispSize;
        end;
      finally
        Free;
      end;
end;

procedure TNkVMainForm.Resize;
begin
  if Image.Loaded then
    UpdateDispSize;
end;

procedure TNkVMainForm.MouseWheel(Delta: Longint);
begin
  if Delta > 0 then
    cmPrevImage
  else
    cmNextImage;
end;

procedure TNkVMainForm.DispatchCommand(Id: Integer);
begin
  case Id of
    IDM_PLUGINSETTING_FIRST .. IDM_PLUGINSETTING_LAST:
      ShowPluginConfig(Plugins[Id - IDM_PLUGINSETTING_FIRST].Path);
  else
    Command.DispatchCommand(Id);
  end;
end;

procedure TNkVMainForm.ShowPluginConfig(const PluginFileName: string);
begin
  with TSpiFilter.Create(nil) do
    try
      LoadPlugin(PluginFileName, False);
      if Handle = 0 then
      begin
        Error(FileName, 'Load failed.');
        Exit;
      end;

      if HasConfigDlg then
        try
          ShowConfigDlg(Self.Handle);
        except
          on E: EImgFilterError do Error(FileName, E.Msg);
        end
      else
       MsgBox(Self.Handle, App.Name,
          FileName+':'#13#10'Configuration dialog is not implemented.');
    finally
      Free;
    end;
end;

procedure TNkVMainform.Paint(const Canvas: TCanvas;
    Left, Top, Width, Height: Longint);
var
  cw, ch: Integer;
begin
  GetClientSize(cw, ch);

  Canvas.BeginDraw;

  if Image.Loaded then
  begin
    Image.Draw(Canvas.Handle, OrgLeft, OrgTop,
      ViewWidth, ViewHeight, ViewLeft, ViewTop,
      DispWidth, DispHeight, ShowGrid and (DispWidth div Image.Width >= 1),
      ScaleFilter, GridSize);

    if ViewWidth < cw then
    begin
      Canvas.FillRect(OrgLeft+ViewWidth, OrgTop,
          cw-(OrgLeft+ViewWidth), ViewHeight, BgColor);
      Canvas.FillRect(0, OrgTop, OrgLeft, ViewHeight, BgColor);
    end;
    if ViewHeight < ch then
    begin
      Canvas.FillRect(0, OrgTop+ViewHeight,
          cw, ch-(OrgTop+ViewHeight), BgColor);
      Canvas.FillRect(0, 0, cw, OrgTop, BgColor);
    end;
  end else
    Canvas.FillRect(0, 0, cw, ch, BgColor);

  Canvas.EndDraw;
end;

procedure TNkVMainForm.MouseDown(const Button: TMouseButton; CX, CY: Longint);
begin
  if Border or FullScreen or (not Border and (mkShift in ModifierKeys)) then
  begin
    SetCapture(True);
    if Image.Loaded then
      Cursor.SetImage(Cursors.Resource('HANDGRAB'));
    OldPt := Cursor.GetPosition;
  end else
  begin
    SetCapture(False);
    StartDrag;
  end;
end;

procedure TNkVMainForm.MouseUp(const Button: TMouseButton; CX, CY: Longint);
begin
  SetCapture(False);
end;

procedure TNkVMainForm.MouseDoubleClick(const Button: TMouseButton;
    CX, CY: Longint);
begin
  cmFitWindow;
end;

procedure TNkVMainForm.MouseMove(CX, CY: Longint);
var
  pt: TPoint;
begin
  if Capture then
  begin
    pt := Cursor.GetPosition;
    Scroll(OldPt.x - pt.x, OldPt.y - pt.y);
    OldPt := pt;
  end
  else if ShowColorValue then
    UpdateCaption;
end;

procedure TNkVMainForm.ShowContextMenu(SX, SY: Longint);
var
  i, id: Integer;
begin
  if mkCtrl in ModifierKeys then
  begin
    if Image.Loaded and (FileList.Items[FileList.Cur] <> '') then
    begin
      ContextMenu := TContextMenu.Create(Handle, FileList.Items[FileList.Cur]);
      ContextMenu.Popup(SX, SY);
      ContextMenu.Free;
      ContextMenu := nil;
    end;
    Exit;
  end;

  with TPopupMenu.Create do
    try
      Command.AddMenuItems(Handle, MenuItems);
      if Image.Loaded then
      begin
        if FileList.Items[FileList.Cur] = '' then
        begin
          GrayItem(Command.Id(cmPrevImage));
          GrayItem(Command.Id(cmNextImage));
          GrayItem(Command.Id(cmPrevFile));
          GrayItem(Command.Id(cmNextFile));
          GrayItem(Command.Id(cmCopyFileName));
          GrayItem(Command.Id(cmReloadImage));
          GrayItem(Command.Id(cmLastFile));
          GrayItem(Command.Id(cmFirstFile));
        end;
      end else
      begin
        GrayItem(Command.Id(cmPrevImage));
        GrayItem(Command.Id(cmNextImage));
        GrayItem(Command.Id(cmPrevFile));
        GrayItem(Command.Id(cmNextFile));
        GrayItem(Command.Id(cmFirstFile));
        GrayItem(Command.Id(cmLastFile));
        GrayItem(Command.Id(cmInfo));
        GrayItem(Command.Id(cmFullScreen));
        GrayItem(Command.Id(cmFitWindow));
        GrayItem(Command.Id(cmZoom50));
        GrayItem(Command.Id(cmZoom100));
        GrayItem(Command.Id(cmZoom150));
        GrayItem(Command.Id(cmZoom200));
        GrayItem(Command.Id(cmZoom300));
        GrayItem(Command.Id(cmZoom400));
        GrayItem(Command.Id(cmZoom800));
        GrayItem(Command.Id(cmZoomIn));
        GrayItem(Command.Id(cmZoomOut));
        GrayItem(Command.Id(cmShowColorValue));
        GrayItem(Command.Id(cmCopyFileName));
        GrayItem(Command.Id(cmReloadImage));
        GrayItem(Command.Id(cmSaveAs));
        GrayItem(Command.Id(cmSetWallpaper));
        GrayItem(Command.Id(cmSetWallpaperTile));
        GrayItem(Command.Id(cmSetImageSize));
        GrayItem(Command.Id(cmOpenDir));
      end; 

      if FullScreen then
      begin
        GrayItem(Command.Id(cmFitWindow));
        GrayItem(Command.Id(cmHideBorder));
      end;

      if Length(Plugins) > 0 then
      begin
        id := Command.Id(cmPluginMenu);
        for i := 0 to High(Plugins) do
          InsertItem(id, BaseName(Plugins[i].Path), IDM_PLUGINSETTING_FIRST+i);
        DeleteItem(id);
      end else
        GrayItem(Command.Id(cmPluginMenu));

      CheckItem(Command.Id(cmFullScreen), FullScreen);
      CheckItem(Command.Id(cmStayOnTop), StayOnTop);
      CheckItem(Command.Id(cmFitImage), FitImage);
      CheckItem(Command.Id(cmFitLargeImage), FitLargeImage);
      CheckItem(Command.Id(cmHideBorder), not Border);
      CheckItem(Command.Id(cmShowColorValue), ShowColorValue);
      CheckItem(Command.Id(cmScaleFilter), ScaleFilter);
      CheckItem(Command.Id(cmShowGrid), ShowGrid);

      Popup(Self.Handle, SX, SY);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.DropFiles(const ListDropFiles: TListDropFiles);
var
  list: TFileList;
  path: WideString;
begin
  list := TFileList.Create;
  try
    while ListDropFiles.Each(path) do
      if IsDir(path) then
      begin
        DirMode := ListDropFiles.Count = 1;
        try
          list.AddDir(path);
        except
          on E: EIOError do Error(E.Path, E.Msg);
        end;
      end else
        list.Add(path);
    list.Sort;
    LoadFirstFile(list);
  finally
    list.Free;
  end;
  if (FileList.Count = 1) and (FileList.Items[0] <> '') then // not stdin 
  begin
    DirMode := True;
    FileListUpdate := True;
  end;
end;

procedure TNkVMainForm.RecvFromRemote(const S: WideString);
var
  list: TFileList;
  args: TWideStringArray;
  i: Integer;
begin
  list := TFileList.Create;
  try
    args := SplitCommandLineW(S);
    for i := 0 to High(args) do
      if IsDir(args[i]) then
      begin
        DirMode := Length(args) = 1;
        try
          list.AddDir(args[i]);
        except
          on E: EIOError do Error(E.Path, E.Msg);
        end;
      end else
        list.Add(args[i]);
    list.Sort;
    Restore;
    LoadFirstFile(list);
  finally
    list.Free;
  end;
  if (FileList.Count = 1) and (FileList.Items[0] <> '') then // not stdin 
  begin
    DirMode := True;
    FileListUpdate := True;
  end;
end;

end.
